from torch.utils.data import DataLoader
from torchnlp.datasets import multi30k_dataset
from data.tokenizer import load_tokenizers, load_vocab
from data.preprocessing import Collator


class TranslateData:
    def __init__(self, config):
        self.config = config

        spacy_de, spacy_en = load_tokenizers()
        self.tokenizer = {"de": spacy_de, "en": spacy_en}

        vocab_src, vocab_tgt = load_vocab(
            self.tokenizer, self.config.vocab_path, self.config.dataset_path
        )
        self.vocab = {"de": vocab_src, "en": vocab_tgt}

        collator = Collator(self.tokenizer, self.vocab, self.config.max_padding)

        train, val, test = multi30k_dataset(
            directory=self.config.dataset_path,
            train=True,
            dev=True,
            test=True,
            urls=[
                "https://raw.githubusercontent.com/neychev/small_DL_repo/master/datasets/Multi30k/training.tar.gz",
                "https://raw.githubusercontent.com/neychev/small_DL_repo/master/datasets/Multi30k/validation.tar.gz",
                "https://raw.githubusercontent.com/neychev/small_DL_repo/master/datasets/Multi30k/mmt16_task1_test.tar.gz",
            ],
        )

        self.train_dataloader = DataLoader(
            train,
            batch_size=self.config.batch_size,
            shuffle=True,
            collate_fn=collator,
        )

        self.val_dataloader = DataLoader(
            val,
            batch_size=self.config.batch_size,
            shuffle=True,
            collate_fn=collator,
        )



from algorithm import SSLTrainer
from model import TransformerModel
from data import TranslateData

import hydra
from omegaconf import DictConfig, OmegaConf
import torch


def setup(config):
    device = torch.device(config.device)
    return device


@hydra.main(version_base=None, config_path="config", config_name="config")
def main(config: DictConfig) -> None:
    ## SETUP ##
    device = setup(config.system)

    ## DATA ##
    data = TranslateData(config.data)
    print("Data Loaded.")

    ## MODEL ##
    src_vocab, tgt_vocab = len(data.vocab["de"]), len(data.vocab["en"])
    model = TransformerModel(src_vocab, tgt_vocab, config.model, device)
    print("Model Created.")

    ## ALGORITHM ##
    print("Running Algorithm...")
    algorithm = SSLTrainer(data, model, config.trainer)
    algorithm.run()
    print("Done!")


if __name__ == "__main__":
    main()



import logging
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm
from algorithm.evaluator import Evaluator

logger = logging.getLogger(__name__)


class RLTrainer:
    def __init__(self, env, buffer, agent, config):
        self.config = config
        self.env = env
        self.agent = agent
        self.buffer = buffer
        self.evaluator = Evaluator(env, agent, config.evaluator)

    def run_training(self):
        env = self.env.env
        obs, info = env.reset()
        actor_loss, critic_loss = 0.0, 0.0
        for step in range(self.buffer.capacity):
            act, logprob = self.agent.act(obs)
            next_obs, reward, terminated, truncated, info = env.step(act)
            done = terminated or truncated
            self.buffer.insert(
                dict(
                    obs=obs,
                    next_obs=next_obs,
                    act=act,
                    logprob=logprob,
                    rew=reward,
                    done=terminated,
                )
            )
            if done:
                obs, info = env.reset()
            else:
                obs = next_obs

        num_updates = self.config.num_updates
        for _ in range(num_updates):
            batch = self.buffer.get_batch()
            loss = self.agent.train(batch)
            actor_loss += loss["actor"]
            critic_loss += loss["critic"]
        self.buffer.reset()

        actor_loss /= num_updates
        critic_loss /= num_updates

        metrics = {"actor_loss": actor_loss, "critic_loss": critic_loss}
        return metrics

    def run_evaluation(self):
        self.evaluator.run()

    def run(self):
        logger.info("Starting training...")
        total_steps = self.config.epochs * self.buffer.capacity
        logger.info(f"Num Epochs: {self.config.epochs}")
        logger.info(f"Num Updates per Epoch: {self.config.num_updates}")
        logger.info(f"Total Steps: {total_steps}")

        with (
            logging_redirect_tqdm(),
            tqdm(
                total=self.config.epochs, desc="Training Progress", unit="epoch"
            ) as progress_bar,
        ):
            for epoch in range(1, self.config.epochs + 1):
                metrics = self.run_training()
                if epoch % self.config.log_interval == 0:
                    actor_loss, critic_loss = (
                        metrics["actor_loss"],
                        metrics["critic_loss"],
                    )
                    logger.debug(
                        f"Epoch {epoch} | Actor Loss: {actor_loss:.4f} | Critic Loss: {critic_loss:.4f}"
                    )
                if epoch % self.config.eval_interval == 0:
                    logger.info(f"Epoch {epoch}")
                    self.run_evaluation()
                progress_bar.update(1)

        self.agent.save()
        logger.info("Training completed.")



from cookiecutter.main import cookiecutter
import click
import os


@click.group()
def main():
    pass


@click.command()
def info():
    click.echo(
        "A modular, best-practice pytorch template for all your deep learning projects."
    )


@click.command()
@click.argument("type", type=click.Choice(["sl", "ssl", "rl"]))
def init(type):
    BASE_PATH = os.path.dirname(__file__) + "/templates"
    cookiecutter(BASE_PATH, directory=type)


main.add_command(info)
main.add_command(init)

if __name__ == "__main__":
    main()



import torch
from torch.nn.functional import pad
from data.tokenizer import tokenize


def subsequent_mask(size):
    # Mask out subsequent positions. Boolean mask
    attn_shape = (1, size, size)
    subsequent_mask = torch.triu(torch.ones(attn_shape), diagonal=1).type(torch.uint8)
    return subsequent_mask == 0


class Batch(object):
    """
    Object for holding a batch of data with mask during training.
    """

    def __init__(self, src, tgt=None, pad=2):  # 2 = <blank>
        self.src = src
        self.src_mask = src == pad
        if tgt is not None:
            self.tgt = tgt[:, :-1]
            self.tgt_y = tgt[:, 1:]
            self.tgt_mask = self.tgt == pad
            self.ntokens = (self.tgt_y != pad).data.sum()

    def to(self, device):
        self.src = self.src.to(device)
        self.tgt = self.tgt.to(device)
        self.src_mask = self.src_mask.to(device)
        self.tgt_mask = self.tgt_mask.to(device)
        self.tgt_y = self.tgt_y.to(device)
        self.ntokens = self.ntokens.to(device)


class Collator(object):
    def __init__(self, tokenizer, vocab, max_padding):
        self.tokenizer = tokenizer
        self.vocab = vocab
        self.max_padding = max_padding

    def __call__(self, batch):
        bs_id = torch.tensor([0])  # <s> token id
        eos_id = torch.tensor([1])  # </s> token id
        pad_id = 2
        src_list, tgt_list = [], []
        for s in batch:
            _src, _tgt = s["de"], s["en"]
            processed_src = torch.cat(
                [
                    bs_id,
                    torch.tensor(
                        self.vocab["de"](tokenize(_src, self.tokenizer["de"])),
                        dtype=torch.int64,
                    ),
                    eos_id,
                ],
                0,
            )
            processed_tgt = torch.cat(
                [
                    bs_id,
                    torch.tensor(
                        self.vocab["en"](tokenize(_tgt, self.tokenizer["en"])),
                        dtype=torch.int64,
                    ),
                    eos_id,
                ],
                0,
            )
            src_list.append(
                # warning - overwrites values for negative values of padding - len
                pad(
                    processed_src,
                    (
                        0,
                        self.max_padding - len(processed_src),
                    ),
                    value=pad_id,
                )
            )
            tgt_list.append(
                pad(
                    processed_tgt,
                    (0, self.max_padding - len(processed_tgt)),
                    value=pad_id,
                )
            )

        src = torch.stack(src_list)
        tgt = torch.stack(tgt_list)
        return Batch(src, tgt)



import torch
from torch import nn, optim
from model.network import MLP
import logging
import os

logger = logging.getLogger(__name__)


class ClassifierModel:
    def __init__(self, config, device):
        logger.info("Creating model...")
        self.config = config
        self.device = device

        self.network = MLP(
            self.config.network.input_dim,
            self.config.network.output_dim,
            self.config.network.hidden_dims,
        ).to(self.device)

        self.optimizer = optim.Adam(
            self.network.parameters(),
            lr=self.config.optimizer.learning_rate,
            weight_decay=self.config.optimizer.weight_decay,
        )
        self.loss = nn.CrossEntropyLoss()

        self.scheduler = optim.lr_scheduler.ReduceLROnPlateau(
            self.optimizer,
            mode="min",
            factor=self.config.scheduler.factor,
            patience=self.config.scheduler.patience,
        )

        if self.config.load_from_checkpoint:
            self.load()

        num_trainable_params = sum(
            p.numel() for p in self.network.parameters() if p.requires_grad
        )
        num_total_params = sum(p.numel() for p in self.network.parameters())
        logger.info(
            f"Trainable Parameters: {num_trainable_params:,} of {num_total_params:,} ({num_trainable_params / num_total_params:.2%})"
        )
        logger.info("Model created.")

    def train_mode(self):
        self.network.train(True)

    def eval_mode(self):
        self.network.train(False)

    def predict(self, image):
        image = image.to(self.device)
        with torch.no_grad():
            pred = self.network(image)
        return pred

    def train(self, batch, grad_accum_steps=1):
        images, labels = batch
        images, labels = images.to(self.device), labels.to(self.device)
        preds = self.network(images)
        loss = self.loss(preds, labels) / grad_accum_steps
        loss.backward()
        return loss.item() * grad_accum_steps

    def validate(self, batch):
        images, labels = batch
        images, labels = images.to(self.device), labels.to(self.device)
        with torch.no_grad():
            preds = self.network(images)
        loss = self.loss(preds, labels).item()
        accuracy = (
            preds.argmax(1) == labels
        ).clone().detach().sum().item() / labels.size(0)
        return loss, accuracy

    def optimizer_step(self):
        self.optimizer.step()
        self.optimizer.zero_grad()

    def scheduler_step(self, val_loss):
        self.scheduler.step(val_loss)

    def save(self):
        save_path = self.config.checkpoint_path
        logger.info(f"Saving model to {save_path}")
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        torch.save(self.network.state_dict(), save_path)

    def load(self):
        load_path = self.config.checkpoint_path
        if os.path.exists(load_path):
            logger.info(f"Loading model from {load_path}")
            self.network.load_state_dict(
                torch.load(load_path, map_location=self.device)
            )
        else:
            logger.error(f"Checkpoint path not found: {load_path}")
            raise FileNotFoundError(f"Checkpoint path not found: {load_path}")



import logging
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

logger = logging.getLogger(__name__)


class SLTrainer:
    def __init__(self, data, model, config):
        self.model = model
        self.data = data
        self.config = config

    def run_validation(self):
        self.model.eval_mode()
        val_loss = 0.0
        val_accuracy = 0.0
        steps_per_epoch = len(self.data.val_dataloader)

        for _, batch in enumerate(self.data.val_dataloader):
            loss, accuracy = self.model.validate(batch)
            val_loss += loss
            val_accuracy += accuracy

        val_loss /= steps_per_epoch
        val_accuracy /= steps_per_epoch
        metrics = {"loss": val_loss, "accuracy": val_accuracy}
        return metrics

    def run_training(self, epoch, progress_bar):
        self.model.train_mode()
        steps_per_epoch = len(self.data.train_dataloader)
        training_loss = 0.0
        for step, batch in enumerate(self.data.train_dataloader, start=1):
            loss = self.model.train(
                batch, grad_accum_steps=self.config.grad_accum_steps
            )
            training_loss += loss

            if step % self.config.grad_accum_steps == 0:
                self.model.optimizer_step()
                progress_bar.update(1)
                current_epoch_fraction = epoch - 1 + (step / steps_per_epoch)
                progress_bar.set_postfix(
                    {"epoch": f"{current_epoch_fraction:.2f}/{self.config.epochs:.2f}"}
                )

            if step % self.config.log_interval == 0:
                logger.debug(
                    f"Epoch {epoch} | Step {step:04d}/{steps_per_epoch} | Training Loss: {loss:.4f}"
                )

        training_loss /= steps_per_epoch
        metrics = {"loss": training_loss}
        return metrics

    def run(self):
        logger.info("Starting training...")
        total_steps = (
            self.config.epochs
            * len(self.data.train_dataloader)
            // self.config.grad_accum_steps
        )
        effective_batch_size = (
            self.data.train_dataloader.batch_size * self.config.grad_accum_steps
        )
        logger.info(f"Num Epochs: {self.config.epochs}")
        logger.info(f"Gradient Accumulation: {self.config.grad_accum_steps}")
        logger.info(f"Effective Batch Size: {effective_batch_size}")
        logger.info(f"Total Steps: {total_steps}")

        with (
            logging_redirect_tqdm(),
            tqdm(
                total=total_steps, desc="Training Progress", unit="step"
            ) as progress_bar,
        ):
            for epoch in range(1, self.config.epochs + 1):
                logger.info(f"Epoch {epoch}")
                train_metrics = self.run_training(epoch, progress_bar)
                train_loss = train_metrics["loss"]
                logger.info(f"Training Loss: {train_loss:>8f}")
                val_metrics = self.run_validation()
                val_loss, val_accuracy = val_metrics["loss"], val_metrics["accuracy"]
                logger.info(
                    f"Validation Loss: {val_loss:>8f}, Validation Accuracy: {(100.0 * val_accuracy):>0.1f}%"
                )
                self.model.scheduler_step(val_loss)

        self.model.save()
        logger.info("Training completed.")



import spacy
import os
import torch
from os.path import exists

from torchtext.vocab import build_vocab_from_iterator
from torchnlp.datasets import multi30k_dataset


def load_tokenizers():
    spacy_de = spacy.load("de_core_news_sm")
    spacy_en = spacy.load("en_core_web_sm")
    return spacy_de, spacy_en


def tokenize(text, tokenizer):
    return [tok.text for tok in tokenizer.tokenizer(text)]


def yield_tokens(data_iter, tokenizer, index):
    for from_to_tuple in data_iter:
        yield tokenizer(from_to_tuple[index])


def build_vocabulary(spacy_de, spacy_en, dataset_path):
    def tokenize_de(text):
        return tokenize(text, spacy_de)

    def tokenize_en(text):
        return tokenize(text, spacy_en)

    train, val, test = multi30k_dataset(
        directory=dataset_path,
        train=True,
        dev=True,
        test=True,
        urls=[
            "https://raw.githubusercontent.com/neychev/small_DL_repo/master/datasets/Multi30k/training.tar.gz",
            "https://raw.githubusercontent.com/neychev/small_DL_repo/master/datasets/Multi30k/validation.tar.gz",
            "https://raw.githubusercontent.com/neychev/small_DL_repo/master/datasets/Multi30k/mmt16_task1_test.tar.gz",
        ],
    )

    print("Building German Vocabulary ...")
    vocab_src = build_vocab_from_iterator(
        yield_tokens(train + val + test, tokenize_de, index="de"),
        min_freq=2,
        specials=["<s>", "</s>", "<blank>", "<unk>"],
    )

    print("Building English Vocabulary ...")
    vocab_tgt = build_vocab_from_iterator(
        yield_tokens(train + val + test, tokenize_en, index="en"),
        min_freq=2,
        specials=["<s>", "</s>", "<blank>", "<unk>"],
    )

    vocab_src.set_default_index(vocab_src["<unk>"])
    vocab_tgt.set_default_index(vocab_tgt["<unk>"])

    return vocab_src, vocab_tgt


def load_vocab(tokenizers, vocab_path, dataset_path):
    spacy_de, spacy_en = tokenizers["de"], tokenizers["en"]
    if not exists(vocab_path):
        vocab_src, vocab_tgt = build_vocabulary(spacy_de, spacy_en, dataset_path)
        torch.save((vocab_src, vocab_tgt), vocab_path)
    else:
        vocab_src, vocab_tgt = torch.load(vocab_path)
    print("Vocabulary sizes:")
    print(len(vocab_src))
    print(len(vocab_tgt))
    return vocab_src, vocab_tgt



import gymnasium as gym
import logging

logger = logging.getLogger(__name__)


class GymEnvironment:
    def __init__(self, config):
        logger.info("Building Gym Environment...")
        self.config = config
        self.env = gym.make(
            self.config.name, max_episode_steps=self.config.max_ep_steps
        )
        self.obs_space = self.env.observation_space
        self.act_space = self.env.action_space
        self.max_ep_steps = self.config.max_ep_steps
        logger.info(f"Environment Name: {self.config.name}")
        logger.info(f"Max Episode Steps: {self.max_ep_steps}")
        logger.info("Gym Environment Built.")



import torch
from torch import nn
import torch.nn.functional as F


class MLP(nn.Module):
    def __init__(self, input_dim, output_dim, hidden_dims):
        super().__init__()
        layers = []
        current_dim = input_dim
        for hidden_dim in hidden_dims:
            layers.append(nn.Linear(current_dim, hidden_dim))
            layers.append(nn.ReLU())
            current_dim = hidden_dim
        layers.append(nn.Linear(current_dim, output_dim))
        self.network = nn.Sequential(*layers)

    def forward(self, x):
        return self.network(x)


class CategoricalActor(nn.Module):
    def __init__(self, obs_dim, act_dim, hidden_dims):
        super().__init__()
        self.network = MLP(obs_dim, act_dim, hidden_dims)

    def forward(self, obs):
        x = self.network(obs)
        probs = F.softmax(x, dim=-1)
        dist = torch.distributions.Categorical(probs)
        return dist


class Critic(nn.Module):
    def __init__(self, obs_dim, hidden_dims):
        super().__init__()
        self.network = MLP(obs_dim, 1, hidden_dims)

    def forward(self, obs):
        x = self.network(obs)
        return x



import os

# Remove LICENSE if "No license file"
if "{{ cookiecutter.license }}" == "None":
    os.remove("LICENSE")

# Choose lockfile based on environment manager
if "{{ cookiecutter.environment_manager }}" == "uv":
    os.remove("requirements.txt")
else:
    os.remove("uv.lock")



import logging
from gymnasium.wrappers import RecordEpisodeStatistics

logger = logging.getLogger(__name__)


class Evaluator:
    def __init__(self, env, agent, config):
        self.config = config
        self.agent = agent
        self.env = env

    def run(self):
        env = RecordEpisodeStatistics(
            self.env.env, buffer_length=self.config.eval_episodes
        )

        for _ in range(self.config.eval_episodes):
            obs, info = env.reset()
            while True:
                act, _ = self.agent.act(obs)
                obs, reward, terminated, truncated, info = env.step(act)
                done = terminated or truncated
                if done:
                    break

        avg_return = sum(list(env.return_queue)) / len(env.return_queue)
        avg_length = sum(list(env.length_queue)) / len(env.length_queue)

        logger.info(f"Evaluation over {self.config.eval_episodes} episodes.")
        logger.info(
            f"Average Return: {avg_return:.2f}, Average Length: {avg_length:.2f}"
        )



from torchvision import datasets
from torchvision.transforms import ToTensor
from torch.utils.data import DataLoader
import logging

logger = logging.getLogger(__name__)


class FashionMNISTData:
    def __init__(self, config):
        logger.info("Loading data...")
        self.config = config
        # Datasets
        self.train_dataset = datasets.FashionMNIST(
            root=self.config.dataset_path,
            train=True,
            download=True,
            transform=ToTensor(),
        )
        logger.info(f"Train Dataset Size: {len(self.train_dataset)}")

        self.val_dataset = datasets.FashionMNIST(
            root=self.config.dataset_path,
            train=False,
            download=True,
            transform=ToTensor(),
        )
        logger.info(f"Validation Dataset Size: {len(self.val_dataset)}")

        # Dataloaders
        batch_size = self.config.batch_size
        self.train_dataloader = DataLoader(self.train_dataset, batch_size)
        self.val_dataloader = DataLoader(self.val_dataset, batch_size)
        logger.info(f"Batch Size: {self.config.batch_size}")
        logger.info("Data loaded.")



import torch


class Batch:
    def __init__(self, data, batch_size, minibatch_size):
        self.data = {key: torch.tensor(val) for key, val in data.items()}
        self.batch_size = batch_size
        self.minibatch_size = minibatch_size
        if self.batch_size % self.minibatch_size != 0:
            raise ValueError(
                f"batch_size ({self.batch_size}) must be divisible by "
                f"minibatch_size ({self.minibatch_size})"
            )
        self.num_minibatches = self.batch_size // self.minibatch_size

    def __iter__(self):
        batch_inds = torch.randperm(self.batch_size)
        for start in range(0, self.batch_size, self.minibatch_size):
            end = start + self.minibatch_size
            minibatch_inds = batch_inds[start:end]
            minibatch = dict()
            for key in self.data:
                minibatch[key] = self.data[key][minibatch_inds]
            yield minibatch



from algorithm import SLTrainer
from model import ClassifierModel
from data import FashionMNISTData

import logging
import torch
import hydra
from omegaconf import DictConfig

logger = logging.getLogger(__name__)


def setup(config):
    """Set up the environment."""
    torch.manual_seed(42)
    device = torch.device(config.device)
    logger.info(f"Device: {device}")
    return device


def cleanup():
    pass


@hydra.main(version_base=None, config_path="config", config_name="config")
def main(config: DictConfig) -> None:
    """Main function to run the training pipeline."""
    # --- SETUP ---
    device = setup(config.system)

    # --- DATA ---
    data = FashionMNISTData(config.data)

    # --- MODEL ---
    model = ClassifierModel(config.model, device)

    # --- ALGORITHM ---
    trainer = SLTrainer(data, model, config.trainer)
    trainer.run()

    # --- CLEANUP ---
    cleanup()


if __name__ == "__main__":
    main()



import torch
import torch.nn.functional as F
from torch import optim
from agent.network import CategoricalActor, Critic
import logging
import os

logger = logging.getLogger(__name__)


class PPOAgent:
    def __init__(self, obs_space, act_space, config, device):
        logger.info("Creating agent...")
        self.config = config
        self.device = device
        self.obs_dim, self.act_dim = obs_space.shape[0], act_space.n
        logger.info(f"Observation Dimension: {self.obs_dim}")
        logger.info(f"Action Dimension: {self.act_dim}")

        self.actor = CategoricalActor(
            self.obs_dim, self.act_dim, self.config.actor.hidden_dims
        ).to(device)
        self.actor_optimizer = optim.Adam(
            self.actor.parameters(), lr=self.config.actor.optimizer.learning_rate
        )

        self.critic = Critic(self.obs_dim, self.config.critic.hidden_dims).to(device)
        self.critic_optimizer = optim.Adam(
            self.critic.parameters(), lr=self.config.critic.optimizer.learning_rate
        )
        logger.info("Agent created.")

    def act(self, obs):
        obs = torch.from_numpy(obs).unsqueeze(0).to(self.device)
        with torch.no_grad():
            dist = self.actor(obs)
            action = dist.sample()
            logprob = dist.log_prob(action)
        action, logprob = (
            action.squeeze(0).detach().cpu().numpy(),
            logprob.squeeze(0).detach().cpu().numpy(),
        )
        return action, logprob

    def gae_estimation(self, batch):
        obs, next_obs = batch["obs"], batch["next_obs"]
        done = batch["done"]
        rew = batch["rew"]

        batch_size = len(obs)
        chunk_size = self.config.gae_chunk_size or batch_size

        values = torch.zeros(batch_size, 1)
        next_values = torch.zeros(batch_size, 1)

        with torch.no_grad():
            for start in range(0, batch_size, chunk_size):
                end = min(start + chunk_size, batch_size)
                obs_chunk = obs[start:end].to(self.device)
                next_obs_chunk = next_obs[start:end].to(self.device)
                values[start:end] = self.critic(obs_chunk).cpu()
                next_values[start:end] = self.critic(next_obs_chunk).cpu()

        # GAE estimate using a reverse loop
        deltas = rew + self.config.gamma * (~done) * next_values - values
        advantages = torch.zeros_like(deltas)
        advantage = 0
        for t in reversed(range(batch_size)):
            advantage = deltas[t] + self.config.gamma * self.config.gae_lambda * advantage * ~done[t]
            advantages[t] = advantage

        returns = advantages + values
        advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-7)
        return advantages, returns

    def loss(self, batch):
        obs, act, logprob = (
            batch["obs"].to(self.device),
            batch["act"].to(self.device),
            batch["logprob"].to(self.device),
        )
        advantages, returns = (
            batch["advantages"].to(self.device),
            batch["returns"].to(self.device),
        )

        dists = self.actor(obs)
        logprobs, old_logprobs = dists.log_prob(act).unsqueeze(-1), logprob

        ratios = torch.exp(logprobs - old_logprobs)
        surr1 = ratios * advantages
        surr2 = torch.clamp(ratios, 1 - self.config.epsilon, 1 + self.config.epsilon) * advantages
        actor_loss = -torch.min(surr1, surr2).mean()

        values = self.critic(obs)
        critic_loss = F.mse_loss(values, returns)

        return actor_loss, critic_loss

    def train(self, batch):
        advantages, returns = self.gae_estimation(batch.data)
        batch.data["advantages"] = advantages
        batch.data["returns"] = returns

        avg_actor_loss, avg_critic_loss = 0.0, 0.0
        for minibatch in batch:
            actor_loss, critic_loss = self.loss(minibatch)

            self.actor_optimizer.zero_grad()
            self.critic_optimizer.zero_grad()
            loss = actor_loss + self.config.critic_coef * critic_loss
            loss.backward()
            self.actor_optimizer.step()
            self.critic_optimizer.step()

            avg_actor_loss += actor_loss.item()
            avg_critic_loss += critic_loss.item()

        avg_actor_loss /= batch.num_minibatches
        avg_critic_loss /= batch.num_minibatches

        losses = {"actor": avg_actor_loss, "critic": avg_critic_loss}
        return losses

    def save(self):
        save_path = self.config.checkpoint_path
        logger.info(f"Saving model to {save_path}")
        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        torch.save(
            {
                "actor_state_dict": self.actor.state_dict(),
                "critic_state_dict": self.critic.state_dict(),
            },
            save_path,
        )

    def load(self):
        load_path = self.config.checkpoint_path
        if os.path.exists(load_path):
            logger.info(f"Loading model from {load_path}")
            checkpoint = torch.load(load_path, map_location=self.device)
            self.actor.load_state_dict(checkpoint["actor_state_dict"])
            self.critic.load_state_dict(checkpoint["critic_state_dict"])
        else:
            logger.error(f"Checkpoint path not found: {load_path}")
            raise FileNotFoundError(f"Checkpoint path not found: {load_path}")



from torch import nn


class MLP(nn.Module):
    def __init__(self, input_dim, output_dim, hidden_dims):
        super().__init__()
        layers = []
        current_dim = input_dim
        for hidden_dim in hidden_dims:
            layers.append(nn.Linear(current_dim, hidden_dim))
            layers.append(nn.ReLU())
            current_dim = hidden_dim
        layers.append(nn.Linear(current_dim, output_dim))
        self.network = nn.Sequential(*layers)
        self.flatten = nn.Flatten()

    def forward(self, images):
        x = self.flatten(images)
        preds = self.network(x)
        return preds



import time


class TrainState:
    """Track number of steps, examples, and tokens processed"""

    step: int = 0  # Steps in the current epoch
    accum_step: int = 0  # Number of gradient accumulation steps
    samples: int = 0  # total # of examples used
    tokens: int = 0  # total # of tokens processed


class SSLTrainer:
    def __init__(self, data, model, config):
        self.data = data
        self.model = model
        self.config = config
        self.train_state = TrainState()

    def run_training(self):
        self.model.set_mode(is_training=True)
        start = time.time()
        total_tokens = 0
        total_loss = 0
        tokens = 0
        n_accum = 0
        train_state = self.train_state

        for i, batch in enumerate(self.data.train_dataloader):
            loss = self.model.train(batch)
            train_state.step += 1
            train_state.samples += batch.src.shape[0]
            train_state.tokens += batch.ntokens
            if i % self.config.accum_interval == 0:
                self.model.optimizer_step()
                self.model.scheduler_step()
                n_accum += 1
                train_state.accum_step += 1
            total_loss += loss
            total_tokens += batch.ntokens
            tokens += batch.ntokens
            if i % self.config.log_interval == 0:
                lr = self.model.get_lr()
                elapsed = time.time() - start
                print(
                    (
                        "Epoch Step: %6d | Accumulation Step: %3d | Loss: %6.2f "
                        + "| Tokens / Sec: %7.1f | Learning Rate: %6.1e"
                    )
                    % (i, n_accum, loss / batch.ntokens, tokens / elapsed, lr)
                )
                start = time.time()
                tokens = 0
        training_loss = total_loss / total_tokens
        print(f"Epoch Loss: {training_loss}")

    def run(self):
        for epoch in range(self.config.epochs):
            self.run_training()



import torch
from torch import nn


class SimpleLossCompute:
    "A simple loss compute and train function."

    def __init__(self, generator, criterion):
        self.generator = generator
        self.criterion = criterion

    def __call__(self, x, y, norm):
        x = self.generator(x)
        sloss = (
            self.criterion(x.contiguous().view(-1, x.size(-1)), y.contiguous().view(-1))
            / norm
        )
        return sloss.data * norm, sloss


class LabelSmoothing(nn.Module):
    # "Implement label smoothing."

    def __init__(self, size, padding_idx, smoothing=0.0):
        super(LabelSmoothing, self).__init__()
        self.criterion = nn.KLDivLoss(reduction="sum")
        self.padding_idx = padding_idx
        self.confidence = 1.0 - smoothing
        self.smoothing = smoothing
        self.size = size
        self.true_dist = None

    def forward(self, x, target):
        assert x.size(1) == self.size
        true_dist = x.data.clone()
        true_dist.fill_(self.smoothing / (self.size - 2))
        true_dist.scatter_(1, target.data.unsqueeze(1), self.confidence)
        true_dist[:, self.padding_idx] = 0
        mask = torch.nonzero(target.data == self.padding_idx)
        if mask.dim() > 0:
            true_dist.index_fill_(0, mask.squeeze(), 0.0)
        self.true_dist = true_dist
        return self.criterion(x, true_dist.clone().detach())



import torch
import math
from torch import nn
from torch.nn.functional import log_softmax


def subsequent_mask(size):
    # Mask out subsequent positions. Boolean mask
    attn_shape = (size, size)
    subsequent_mask = torch.triu(torch.ones(attn_shape), diagonal=1).type(torch.uint8)
    return subsequent_mask == 1


class Embeddings(nn.Module):
    def __init__(self, d_model, vocab):
        super(Embeddings, self).__init__()
        self.lut = nn.Embedding(vocab, d_model)
        self.d_model = d_model

    def forward(self, x):
        return self.lut(x) * math.sqrt(self.d_model)


class PositionalEncoding(nn.Module):
    # "Implement the PE function."

    def __init__(self, d_model, dropout, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2) * -(math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x):
        x = x + self.pe[:, : x.size(1)].requires_grad_(False)
        return self.dropout(x)


class Generator(nn.Module):
    """
    Define standard linear + softmax generation step.
    """

    def __init__(self, d_model, vocab):
        super(Generator, self).__init__()
        self.proj = nn.Linear(d_model, vocab)

    def forward(self, x):
        return log_softmax(self.proj(x), dim=-1)


class Transformer(nn.Module):
    def __init__(
        self,
        src_vocab,
        tgt_vocab,
        d_model,
        nhead,
        num_encoder_layers,
        num_decoder_layers,
        dim_feedforward,
        dropout,
    ):
        super(Transformer, self).__init__()
        self.transformer = nn.Transformer(
            d_model=d_model,
            nhead=nhead,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            batch_first=True,
        )  # first dimension is batch_size
        self.src_embed = nn.Sequential(
            Embeddings(d_model, src_vocab), PositionalEncoding(d_model, dropout)
        )
        self.tgt_embed = nn.Sequential(
            Embeddings(d_model, tgt_vocab), PositionalEncoding(d_model, dropout)
        )
        self.generator = Generator(d_model, tgt_vocab)

    def forward(self, src, tgt, src_mask=None, tgt_mask=None):
        "Take in and process masked src and target sequences."
        causal_mask = subsequent_mask(tgt.size(-1))
        src, tgt = self.src_embed(src), self.tgt_embed(tgt)
        return self.transformer(
            src=src,
            tgt=tgt,
            tgt_mask=causal_mask,
            src_key_padding_mask=src_mask,
            tgt_key_padding_mask=tgt_mask,
        )

    def encode(self, src, src_mask=None):
        src = self.src_embed(src)
        return self.transformer.encoder(src=src, src_key_padding_mask=src_mask)

    def decode(self, mem, tgt, mem_mask=None, tgt_mask=None):
        causal_mask = subsequent_mask(tgt.size(-1))
        tgt = self.tgt_embed(tgt)
        return self.transformer.decoder(
            tgt=tgt,
            memory=mem,
            tgt_mask=causal_mask,
            tgt_key_padding_mask=tgt_mask,
            memory_key_padding_mask=mem_mask,
        )



import torch
from torch import nn
from model.transformer import Transformer
from torch.optim.lr_scheduler import LambdaLR
from model.loss import SimpleLossCompute, LabelSmoothing


def rate(step, model_size, factor, warmup):
    """
    we have to default the step to 1 for LambdaLR function
    to avoid zero raising to negative power.
    """
    if step == 0:
        step = 1
    return factor * (
        model_size ** (-0.5) * min(step ** (-0.5), step * warmup ** (-1.5))
    )


class TransformerModel:
    def __init__(self, src_vocab, tgt_vocab, config, device):
        self.config = config
        self.src_vocab = src_vocab
        self.tgt_vocab = tgt_vocab

        self.device = device
        self.transformer = Transformer(
            src_vocab,
            tgt_vocab,
            d_model=self.config.network.d_model,
            nhead=self.config.network.h,
            num_encoder_layers=self.config.network.N,
            num_decoder_layers=self.config.network.N,
            dim_feedforward=self.config.network.d_ff,
            dropout=self.config.network.dropout,
        ).to(self.device)

        # This was important from their code.
        # Initialize parameters with Glorot / fan_avg.
        for p in self.transformer.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

        criterion = LabelSmoothing(
            tgt_vocab, padding_idx=2, smoothing=self.config.loss.smoothing
        )
        self.loss = SimpleLossCompute(self.transformer.generator, criterion)
        self.optimizer = torch.optim.Adam(
            self.transformer.parameters(),
            lr=self.config.optimizer.lr,
            betas=(0.9, 0.98),
            eps=1e-9,
        )
        self.scheduler = LambdaLR(
            optimizer=self.optimizer,
            lr_lambda=lambda step: rate(
                step,
                model_size=self.config.network.d_model,
                factor=1.0,
                warmup=self.config.optimizer.warmup,
            ),
        )

    def set_mode(self, is_training=True):
        self.transformer.train(is_training)

    def get_lr(self):
        return self.optimizer.param_groups[0]["lr"]

    def predict(self, src, tgt, src_mask, tgt_mask):
        src = src.to(self.device)
        tgt = tgt.to(self.device)
        src_mask = src_mask.to(self.device)
        tgt_mask = tgt_mask.to(self.device)
        return self.transformer(src, tgt, src_mask, tgt_mask)

    def generate(self, src, src_mask, max_len, start_symbol):
        src = src.to(self.device)
        src_mask = src_mask.to(self.device)
        memory = self.transformer.encode(src, src_mask)
        ys = torch.zeros(1, 1).fill_(start_symbol).type_as(src.data)
        for i in range(max_len - 1):
            out = self.transformer.decode(memory, ys, src_mask)
            prob = self.transformer.generator(out[:, -1])
            _, next_word = torch.max(prob, dim=1)
            next_word = next_word.data[0]
            ys = torch.cat(
                [ys, torch.zeros(1, 1).type_as(src.data).fill_(next_word)], dim=1
            )
        return ys

    def train(self, batch):
        batch.to(self.device)
        pred = self.transformer(batch.src, batch.tgt, batch.src_mask, batch.tgt_mask)
        loss, loss_node = self.loss(pred, batch.tgt_y, batch.ntokens)
        loss_node.backward()
        return loss

    def optimizer_step(self):
        self.optimizer.step()
        self.optimizer.zero_grad(set_to_none=True)

    def scheduler_step(self):
        self.scheduler.step()



from data.batch import Batch
import numpy as np
import logging

logger = logging.getLogger(__name__)


class RolloutBuffer:
    def __init__(self, obs_space, act_space, config):
        logger.info("Initializing Rollout Buffer...")
        self.config = config
        self.capacity = self.config.capacity
        self.minibatch_size = self.config.minibatch_size
        self.num_minibatches = self.capacity // self.minibatch_size
        assert self.capacity % self.minibatch_size == 0, (
            "Buffer capacity must be divisible by the minibatch size."
        )

        self.data = dict(
            obs=np.empty(
                (self.capacity, *obs_space.shape),
                dtype=obs_space.dtype,
            ),
            next_obs=np.empty(
                (self.capacity, *obs_space.shape),
                dtype=obs_space.dtype,
            ),
            act=np.empty((self.capacity, *act_space.shape), dtype=act_space.dtype),
            logprob=np.empty((self.capacity, 1), dtype=np.float32),
            rew=np.empty((self.capacity, 1), dtype=np.float32),
            done=np.empty((self.capacity, 1), dtype=bool),
        )
        self.size = 0
        logger.info(f"Buffer Capacity: {self.capacity}")
        logger.info(f"Minibatch Size: {self.minibatch_size}")
        logger.info(f"Num Minibatches: {self.num_minibatches}")
        logger.info("Empty Buffer Initialized.")

    def insert(self, data):
        assert self.size < self.capacity, "Rollout Buffer is full."
        for key in data:
            self.data[key][self.size] = data[key]
        self.size += 1

    def get_batch(self):
        assert self.size == self.capacity, "Rollout Buffer is not full yet."
        return Batch(
            data=self.data, batch_size=self.capacity, minibatch_size=self.minibatch_size
        )

    def reset(self):
        self.size = 0



from environment import GymEnvironment
from data import RolloutBuffer
from agent import PPOAgent
from algorithm import RLTrainer

import torch
import hydra
from omegaconf import DictConfig


def setup(config):
    torch.manual_seed(42)
    device = torch.device(config.device)
    return device


def cleanup():
    pass


@hydra.main(version_base=None, config_path="config", config_name="config")
def main(config: DictConfig) -> None:
    # --- SETUP ---
    device = setup(config.system)

    # --- ENVIRONMENT ---
    env = GymEnvironment(config.environment)

    # --- DATA ---
    buffer = RolloutBuffer(env.obs_space, env.act_space, config.buffer)

    # --- AGENT ---
    agent = PPOAgent(env.obs_space, env.act_space, config.agent, device)

    # --- ALGORITHM ---
    trainer = RLTrainer(env, buffer, agent, config.trainer)
    trainer.run()

    # --- CLEANUP ---
    cleanup()


if __name__ == "__main__":
    main()



import subprocess
import os

if "{{ cookiecutter.documentation }}" == "mkdocs":
    command = ['mkdocs', 'new', '.', '--quiet']
    subprocess.run(command,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL)
elif "{{ cookiecutter.documentation }}" == "sphinx":
    os.mkdir('docs/')
    command = [
        'sphinx-quickstart',
        '-q',
        '-p', "{{ cookiecutter.project_name }}",
        '-a', "{{ cookiecutter.author_name }}",
        '--no-makefile',
        '--no-batchfile',
        'docs/'
    ]
    subprocess.run(command,
        stdout=subprocess.DEVNULL,  # Suppress standard output
        stderr=subprocess.DEVNULL)
